@@ -1,9 +1,15 @@
+/*
+ * $Log: arg.cpp,v $
+ * Revision 1.3  2005/07/04 21:59:42  martin
+ * added logging to all files
+ *
+ */
 #include "arg.h"
 #include "string_utils.h"
 #include "debug.h"
 
 
-#define OUT(level) DEBUG_OUT( level, MODULE_ARG )
+#define OUT(level) DEBUG_OUT( level, MODULE::ARG )
 
 #include <iostream>
 
@@ -20,7 +26,7 @@ int Arg::get_terminal_width()
     
 #ifdef HAVE_CURSES_H
     
-    const char *tm  = termname();
+    const char *tm  = getenv( "TERM" );
     
     if( !tm || std::string( tm ).empty() )
 	return DEFAULT_COLS;
@@ -491,7 +497,7 @@ void Arg::EnumOption::addEnum( const std::string &name )
   convert<IsValueRule>( value_rule )->getStates()->push_back( name );
 }
 
-Arg::OptionChain::OptionChain( LinkMode m )
+Arg::OptionChain::OptionChain( LINK m )
   : mode( m )
 {
   min_match = 0;
@@ -661,7 +667,7 @@ bool Arg::OptionChain::parse( list_soption &arg )
 
 	} // for
 
-      if( !found && ( mode == LINK_AND ) )
+      if( !found && ( mode == LINK::AND ) )
 	return false;
 
     }  
