@@ -1,10 +1,28 @@
+/*
+  Format a C++ library for typesafe string formating in printf style
+      (C) 2001 - 2003 by Martin Oberzalek <kingleo@gmx.at>  
+
+   2004-01-01 remerged version 1.1 of Format with xstow's Format
+
+  Examples:
+     std::cout << format( "Hello %s, I have $05d$ in my pocket", "world", 5 ) << std::endl;
+     std::cout << format( "Do not try this with printf: %s", 10101 ) << std::endl;
+
+  
+*/
+/*
+ * $Log: format.h,v $
+ * Revision 1.4  2010/07/21 19:38:25  martin
+ * gcc-4 Port
+ *
+ * Revision 1.3  2005/07/04 21:59:42  martin
+ * added logging to all files
+ *
+ */
+
 #ifndef format_h
 #define format_h
 
-/*
-  Experimental Version of Format 2
-*/
-
 #include <string>
 #include <iomanip>
 #include <iostream>
@@ -19,20 +37,23 @@
 #  include <sstream>
 #  include <cctype>
 #  define IS_DIGIT( x ) std::isdigit( x )
-#  define strstream stringstream
-#  define STD_ENDS ""
 #else
 extern "C" {
 #  include <ctype.h>
 }
 #  include <strstream>
 #  define IS_DIGIT( x ) isdigit( x )
-#  define STRSTREAM
-#  define STD_ENDS std::ends
 #endif  
 
 #ifndef NFORMAT
 
+namespace Format
+{
+  class CFormat;
+}
+
+inline std::ostream& operator<<( std::ostream& out, Format::CFormat cf );
+
 namespace Format
 {
   typedef std::string::size_type ST;
@@ -41,23 +62,23 @@ namespace Format
   {    
   public:
     
-    typedef enum Adjust
+    enum Adjust
       {
-	LEFT,
-	RIGHT
+		LEFT,
+		RIGHT
       };
     
-    typedef enum Base
+    enum Base
       {
-	OCT,
-	DEC,
-	HEX
+		OCT,
+		DEC,
+		HEX
       };
     
-    typedef enum Floating
+	enum Floating
       {
-	FIXED,
-	SCIENTIFIC
+		FIXED,
+		SCIENTIFIC
       };
     
     bool valid;
@@ -76,7 +97,8 @@ namespace Format
     
     int width;
     int precision;
-    
+    int strlength;
+
     std::string format;
 
   public:    
@@ -87,7 +109,7 @@ namespace Format
   private:    
     void reset();
   };
-  
+
   /****************************************/
   // all the errors that are thrown 
   // are a cause of a mistake with %* or %*m$
@@ -151,15 +173,18 @@ namespace Format
 
     template <class S> std::string x2s( S ss, const CFormat &cf )
     {
+#ifdef HAVE_STL_SSTREAM
+	std::stringstream str;
+	str << cf << ss;
+	std::string st = str.str();
+	return st;
+#else
       std::strstream str;
-      str << cf << ss << STD_ENDS;
+      str << cf << ss << std::ends;
       std::string st = str.str();
-
-#ifdef STRSTREAM
       str.freeze(0);
-#endif
-
       return st;
+#endif
     }
 
   };
@@ -174,13 +199,12 @@ inline std::ostream& operator<<( std::ostream& out, Format::CFormat cf )
   return out;
 }
 
-
 template <class A, class B, class C, class D, class E, class F>
 Format::Format<A,B,C,D,E,F>::Format( std::string const &format, A a, B b, C c, D d, E e, F f, unsigned int num_of_args )
   : format( format ), a(a), b(b), c(c), d(d), e(e), f(f), num_of_args( num_of_args )
 {
   if( num_of_args > 6 )
-    throw Error( "Number of args out of Range" );
+    throw Error( "Number of args out of range" );
 
   gen_arg_list();
   
@@ -191,7 +215,7 @@ template <class A, class B, class C, class D, class E, class F>
 int Format::Format<A,B,C,D,E,F>::get_int_arg( unsigned int num )
 {
   if( static_cast<unsigned int>(num) > num_of_args - 1 )
-    throw Error( "The arg you wan't to use is out of Range" );
+    throw Error( "The arg you wan't to use is out of range" );
 
   if( num < 0 )
     throw Error( "negativ number for arg number not allowed" );
@@ -283,7 +307,7 @@ void Format::Format<A,B,C,D,E,F>::parse()
   ST pos = 0;
   ST len = format.size();
   s = "";
-  bool had_precisision = false;
+  bool had_precision = false;
 
   while( par < num_of_args && pos < len )
     { // while
@@ -404,10 +428,10 @@ void Format::Format<A,B,C,D,E,F>::parse()
 	      if( !(pos < len) )
 		return;
 	      
-	      had_precisision = true;
+	      had_precision = true;
 
 	      if( IS_DIGIT( format[pos] ) )
-		f.precision = skip_atoi( format, pos, pos );
+		  f.precision = skip_atoi( format, pos, pos );
 	      else if( format[pos] == '*' )
 		{
 		  pos++;
@@ -567,10 +591,19 @@ void Format::Format<A,B,C,D,E,F>::parse()
 	  if( use_par != par )
 	    upar = use_par;
 
+	  if( f.base == CFormat::HEX && had_precision && f.special )
+	  {
+	      CFormat f2;
+	      f2.base = f.base;
+	      std::string s = use_arg( upar, f2 );
+	      f.strlength = s.size();
+//	      printf( "str: %s\n", s.c_str() );
+	  }
+
 	  str = use_arg( upar, f );
 
 	  // cut string
-	  if( had_precisision && args[upar].is_string )
+	  if( had_precision && args[upar].is_string )
 	    str = str.substr( 0, f.precision );
 
 	  s += str;
@@ -603,7 +636,9 @@ void Format::Format<A,B,C,D,E,F>::parse()
 #else // ifndef NFORMAT
 
 #include "string_utils.h" // for x2s
+
 #include <cstdio>
+#include <cstring>
 
 namespace Format
 {
@@ -783,6 +818,15 @@ namespace Format
 	return *this;
       }
 
+    PrintF& operator()( std::string fs )
+    {
+      if( check() )
+	out << fs;
+
+      return *this;
+    }
+
+
     template<typename A> PrintF& operator()( std::string fs, const A &a )
     {
       if( check() )
@@ -860,8 +904,5 @@ namespace Format
 
 
 #undef IS_DIGIT
-#undef strstream
-#undef STRSTREAM
-#undef STD_ENDS
 
 #endif
