@@ -1,4 +1,14 @@
+/*
+ * $Log: main.cpp,v $
+ * Revision 1.3  2010/07/21 19:38:25  martin
+ * gcc-4 Port
+ *
+ * Revision 1.2  2005/07/04 21:59:42  martin
+ * added logging to all files
+ *
+ */
 #include <iostream>
+#include <stdlib.h>
 
 #include "local_config.h"
 
@@ -13,10 +23,14 @@
 
 using namespace std;
 
-#define OUT( level ) DEBUG_OUT( level, MODULE_MAIN )
+#define OUT( level ) DEBUG_OUT( level, MODULE::MAIN )
 
 int main( int argc, char** argv )
 {
+    BT;
+
+    progname = "xstow";
+
   char *dl = getenv( "XSTOW_DEBUG_LEVEL" );
   char *dm = getenv( "XSTOW_DEBUG_MODULE" );
   char *uc = getenv( "XSTOW_USE_CURSES" );
@@ -81,7 +95,7 @@ int main( int argc, char** argv )
 #endif
 
 #ifdef CAN_USE_INI
-  Arg::FileOption files( "F" );
+  Arg::FileOption files( "F", Arg::FILE::ANY );
   files.addName( "file" );
   files.setDescription( "Read this config file too");
   arg.addOptionR( &files );
@@ -100,7 +114,7 @@ int main( int argc, char** argv )
   
   Arg::FlagOption version( "V" );
   version.addName( "Version" );
-  version.setDescription( "Show XStow version number" );
+  version.setDescription( "Show XStow's version number" );
   info.addOptionR( &version );
 
   Arg::OptionChain oc;
@@ -124,12 +138,12 @@ int main( int argc, char** argv )
   simulate.setDescription( "Simulate installation (ignore permissions)" );
   oc.addOptionR( &simulate );
 
-  Arg::FileOption working_dir( "d", Arg::FILE_DIR );
+  Arg::FileOption working_dir( "d", Arg::FILE::DIR );
   working_dir.addName( "dir" );
   working_dir.setDescription( "Set stow dir to DIR (default is current dir)" );
   oc.addOptionR( &working_dir );
 
-  Arg::FileOption target_dir( "t", Arg::FILE_DIR );
+  Arg::FileOption target_dir( "t", Arg::FILE::DIR );
   target_dir.addName( "target" );
   target_dir.setDescription( "Set target to DIR (default is parent of stow dir)" );
   oc.addOptionR( &target_dir );
@@ -240,6 +254,20 @@ int main( int argc, char** argv )
   oc.addOptionR( &copy );
 #endif
 
+  Arg::StringOption ignore_file_in_dir( "ifd", "DIR/FILE" );
+  ignore_file_in_dir.addName( "i-file-in-dir" );
+  ignore_file_in_dir.setDescription( "Ignore this file in this directory. (Relative to the package dir.)" );
+  ignore_file_in_dir.setMinValues( 1 );
+  ignore_file_in_dir.setMaxValues( 1 );
+  oc.addOptionR( &ignore_file_in_dir );
+
+  Arg::StringOption copy_file_in_dir( "cfd", "DIR/FILE" );
+  copy_file_in_dir.addName( "c-file-in-dir" );
+  copy_file_in_dir.setDescription( "Copy this file in this directory. (Relative to the package dir.)" );
+  copy_file_in_dir.setMinValues( 1 );
+  copy_file_in_dir.setMaxValues( 1 );
+  oc.addOptionR( &copy_file_in_dir );
+
 #ifdef HAVE_REGEX_H
   Arg::StringOption copy_regex( "cre", "REGEX" );
   copy_regex.addName( "copy-regex" );
@@ -250,7 +278,7 @@ int main( int argc, char** argv )
 #endif
 
 
-  Arg::StringOption stow_dirs( "sd", "DIRS" );
+  Arg::FileOption stow_dirs( "sd", Arg::FILE::DIR );
   stow_dirs.addName( "stow-dirs" );
   stow_dirs.setDescription( "inform xstow about other stow directories" );
   stow_dirs.setMinValues( 1 );
@@ -275,7 +303,7 @@ int main( int argc, char** argv )
   oc.addOptionR( &stow_dirs_add_regex );
 #endif
 
-  Arg::StringOption protect_dirs( "pd", "DIRS" );
+  Arg::FileOption protect_dirs( "pd", Arg::FILE::DIR );
   protect_dirs.addName( "protect-dirs" );
   protect_dirs.setDescription( "Never install anything in this directory" );
   protect_dirs.setMinValues( 1 );
@@ -317,7 +345,7 @@ int main( int argc, char** argv )
   force.setDescription( "Skip conflicts if possible." );
   oc.addOptionR( &force );
 
-//  Arg::EmptyFileOption packages( Arg::FILE_DIR );
+//  Arg::EmptyFileOption packages( Arg::FILE::DIR );
   Arg::EmptyStringOption packages;
   packages.setRequired( true );
   packages.setMinValues( 1 );
@@ -332,7 +360,7 @@ int main( int argc, char** argv )
 
       if( version.getState() )
 	{
-	  out( "XStow Version %s (C) 2002-2003 by Martin Oberzalek\n", VERSION );
+	  out( "XStow Version %s (C) 2002-2005 by Martin Oberzalek\n", VERSION );
 
 #ifndef NDEBUG
 	  out << "\t advanced debugging support\n";
@@ -373,6 +401,7 @@ int main( int argc, char** argv )
 	}
 
       out << "Usage: xstow [OPTION ...] PACKAGE\n"; 
+      out << " -h for help screen\n";
 
       return 1;
     }
@@ -388,7 +417,7 @@ int main( int argc, char** argv )
       setup.debug.module.freeze( true );
     }
   else
-    setup.debug.module( MODULE_ALL );
+    setup.debug.module( MODULE::ALL );
 
 
 
@@ -415,6 +444,9 @@ int main( int argc, char** argv )
   else
     setup.working_dir( CppDir::exec_path );
 
+  // add working dir to stow dirs for matching all cases
+  setup.stow_dirs.dirs += setup.working_dir();
+
   if( target_dir.isSet() )
     setup.target_dir( CppDir::get_full_path( *(target_dir.getValues()->begin()) ) );
   else
@@ -454,7 +486,7 @@ int main( int argc, char** argv )
       // that's not done by the option anymore since we are trying getting 
       // compatible to Stow :(
 
-      Arg::FileOptionRule check_exist( Arg::FILE_DIR );
+      Arg::FileOptionRule check_exist( Arg::FILE::DIR );
 
       std::string current_path = setup.working_dir();
 
@@ -514,10 +546,14 @@ int main( int argc, char** argv )
     setup.matches.copy_regex += split_simple( *copy_regex.getValues()->begin() );
 #endif
 
+  if( copy_file_in_dir.isSet() )
+      setup.matches.copy_file_in_dir += split_simple( *copy_file_in_dir.getValues()->begin() );
 
+  if( ignore_file_in_dir.isSet() )
+      setup.matches.ignore_file_in_dir += split_simple( *ignore_file_in_dir.getValues()->begin() );
 
   if( stow_dirs.isSet() )
-    setup.stow_dirs.dirs += split_simple( *stow_dirs.getValues()->begin() );
+    setup.stow_dirs.dirs += CppDir::get_full_path( *stow_dirs.getValues()->begin() );
 
 #ifdef HAVE_FNMATCH_H
   if( stow_dirs_add.isSet() )
@@ -532,7 +568,7 @@ int main( int argc, char** argv )
 
 
   if( protect_dirs.isSet() )
-    setup.protect_dirs.dirs += split_simple( *protect_dirs.getValues()->begin() );
+    setup.protect_dirs.dirs += CppDir::get_full_path( *protect_dirs.getValues()->begin() );
 
 #ifdef HAVE_FNMATCH_H
   if( protect_dirs_add.isSet() )
@@ -605,15 +641,20 @@ int main( int argc, char** argv )
 	    Tree tree( setup.packages[i], setup.target_dir(), setup.working_dir() );
 	    
 	    if( tree.scan() )
-	      {     
-		if( out.check( MODULE_MAIN, 1 ) )
-		  out << tree.getTree() << '\n' << std::endl;
+	    {     
+		if( out.check( MODULE::MAIN, 1 ) )
+		    out << tree.getTree() << '\n' << std::endl;
 		
 		if( tree.check() )
-		  if( !tree.exec_list() )
-		    return 3;
-	      }
-	    
+		{
+		    if( !tree.exec_list() )
+			return 3;
+		}
+		else
+		    return 4;
+	    }
+	    else
+		return 4;
 	  }
 
 	  if( again )
@@ -628,12 +669,16 @@ int main( int argc, char** argv )
   catch( std::exception &err )
     {
 	ERROR( "Error: %s\n", err.what() ); 
+#ifndef NDEBUG
+	ERROR( "Backtrace: %s\n", BackTraceHelper::bt.bt() );
+#endif
 	return 2;
     }
 #ifndef NFORMAT
   catch( Format::Error &err )
     {
 	ERROR( "Format::Error: %s\n", err.err ); 
+	return 2;
     }
 #endif
   catch( ... )
@@ -642,5 +687,5 @@ int main( int argc, char** argv )
       return 2;
     }
 
-  return 0;
+ return 0;
 }
