@@ -1,3 +1,14 @@
+/*
+ * $Log: tree.cpp,v $
+ * Revision 1.4  2010/07/21 19:38:25  martin
+ * gcc-4 Port
+ *
+ * Revision 1.3  2005/07/04 21:59:42  martin
+ * added logging to all files
+ *
+ */
+#include <stdlib.h>
+#include <cstring>
 #include "tree.h"
 #include "debug.h"
 #include "iterator.h"
@@ -19,7 +30,6 @@ extern "C" {
 
 #ifdef HAVE_REGEX_H
 extern "C" {
-#  include <sys/types.h>
 #  include <regex.h>
 }
 #endif
@@ -27,7 +37,7 @@ extern "C" {
 
 #include <algorithm>
 
-#define OUT(level) DEBUG_OUT(level, MODULE_TREE)
+#define OUT(level) DEBUG_OUT(level, MODULE::TREE)
 
 namespace {
 
@@ -43,28 +53,30 @@ Node::Node( CppDir::File file, Ref<Node> parent )
     //    fullpath( *this, std::string(), true ),
     path_is_hashed( false )
 {
+    BT;
+
   if( !file )
-    type = UNKNOWN;
+    type = TYPE::UNKNOWN;
 
   switch( file.get_type() )
     {
-    case CppDir::File::DIR:     type = DIR; break;
-    case CppDir::File::FIFO:
-    case CppDir::File::CHAR:
-    case CppDir::File::BLOCK:
-    case CppDir::File::REGULAR: type = FILE; break;
+    case CppDir::EFILE::DIR:     type = TYPE::DIR; break;
+    case CppDir::EFILE::FIFO:
+    case CppDir::EFILE::CHAR:
+    case CppDir::EFILE::BLOCK:
+    case CppDir::EFILE::REGULAR: type = TYPE::FILE; break;
 
     default:
-      type = UNKNOWN;
+      type = TYPE::UNKNOWN;
     }
 
-  if( type == UNKNOWN )
+  if( type == TYPE::UNKNOWN )
     {
       if( file.is_link() )
 	{
 	  OUT(0)( "dead link %s. This does not harm, but can cause problems.\n"
 		  "Handling dead link as normal file.\n", ::getName( file ) );
-	  type = FILE;
+	  type = TYPE::FILE;
 	}
       else
 	OUT(0)( "unknown file type %s\n", ::getName( file ) );
@@ -73,13 +85,15 @@ Node::Node( CppDir::File file, Ref<Node> parent )
 
 bool Node::scan( bool recursion )
 {
+    BT;
+
   if( childs.valid() )
     return true; // already scanned
 
-  if( type == FILE )
+  if( type == TYPE::FILE )
     return true;
 
-  if( type == UNKNOWN )
+  if( type == TYPE::UNKNOWN )
     {
       ERROR( "unknown file type %s\n", ::getName( file ) );
       return false;
@@ -154,6 +168,19 @@ bool Node::scan( bool recursion )
 	  setup.protect_dirs.dirs += ::getName( nn->file.get_name() );
       }
 
+      DEBUG( OUT(2)( "xxxxx file %s type: %s match: %s\n",nn->file.get_name(), nn->getType(), match_ignore( nn->file.get_name() ) ) );
+      
+      if( nn->getType() == TYPE::FILE )
+      {	      
+	  if( match_ignore( nn->file.get_name() ) || 
+	      match_ignore_file_in_dir( setup.current_package(), CppDir::concat_dir( getFullPathName(), nn->file.get_name() ) ) )
+	  {
+	      DEBUG( OUT(2)( "file %s ignored\n", CppDir::concat_dir( getFullPathName(), nn->file.get_name() ) ) );
+	      continue;
+	  }
+	  
+      }	  
+
       childs->push_back( nn );
 
       if( recursion )
@@ -165,7 +192,9 @@ bool Node::scan( bool recursion )
 
 std::string Node::calcFullPathName( const Node &node )
 {
-  if( node.type == UNKNOWN )
+    BT;
+
+  if( node.type == TYPE::UNKNOWN )
     return std::string();
 
   if( node.hasParent() )
@@ -176,6 +205,8 @@ std::string Node::calcFullPathName( const Node &node )
 
 std::string Node::getFullPathName() const
 {
+    BT;
+
   if( !path_is_hashed )
     full_path = calcFullPathName( *this );
 
@@ -184,6 +215,8 @@ std::string Node::getFullPathName() const
 
 Ref<Node> Node::getRoot()
 {
+    BT;
+
   if( !hasParent() )
     return Ref<Node>(*this, false );
 
@@ -195,6 +228,8 @@ Tree::Tree( const std::string &base_,
 	    const std::string &working_
 	    )
 {
+    BT;
+
   base = new Node( CppDir::File( base_ ) );
   target = new Node( CppDir::File( target_ ) );
   working = new Node( CppDir::File( working_ ) );
@@ -203,6 +238,8 @@ Tree::Tree( const std::string &base_,
 
 bool Tree::scan()
 {
+    BT;
+
   DEBUG( OUT(3) << "target path name: " << target->getFullPathName() << std::endl );
 
   bool ret = base->scan();
@@ -214,6 +251,8 @@ bool Tree::scan()
 
 std::ostream& operator<<( std::ostream& out, Ref<Node> node )
 {
+    BT;
+
   if( !node )
     return out;
 
@@ -236,12 +275,14 @@ std::ostream& operator<<( std::ostream& out, Ref<Node> node )
 
 bool Tree::check( Ref<Node> target, Ref<Node> base, Ref<vec_link> links )
 {
+    BT;
+
   CppDir::File file( target->getFile() );
 
   DEBUG( OUT(3) << "target path name: " << target->getFullPathName() << std::endl );
   DEBUG( OUT(3) << "file.get_path(): " << file.get_path() << "\tfile.get_name(): " << file.get_name() << std::endl );
 
-  if( !file || file.get_type() != CppDir::File::DIR )
+  if( !file || file.get_type() != CppDir::EFILE::DIR )
     {
       ERROR( "target dir %s is not a valid directory\n", getName( file ) );
       return false;
@@ -337,7 +378,7 @@ bool Tree::check( Ref<Node> target, Ref<Node> base, Ref<vec_link> links )
 			  
 			  Link link;
 
-			  link.action = Link::REMOVE;
+			  link.action = Link::ACTION::REMOVE;
 			  link.origin = CppDir::concat_dir( dit->get_path(), dit->get_name() );
 
 			  links->push_back( link );
@@ -349,38 +390,62 @@ bool Tree::check( Ref<Node> target, Ref<Node> base, Ref<vec_link> links )
 		  DEBUG( OUT(2) << "\t=> NO\n" );
 
 		}
-	      
-	      if( it->getType() == Node::FILE && match_copy( it->getName() ) &&  CppDir::File::REGULAR )
+
+#ifdef CAN_USE_INI
+	      std::string s_exec;
+	      if( it->getType() == Node::TYPE::FILE && match_exec( it->getName(), s_exec ) )
+	      {
+		  Link link( it->getFullPathName(), 
+			     CppDir::concat_dir( target->getFullPathName(), it->getName() ),
+			     s_exec );
+		  link.action = Link::ACTION::EXEC;
+		  links->push_back( link );
+
+		  continue;
+	      }
+#endif	      
+
+	      if( it->getType() == Node::TYPE::FILE && 
+		  ( match_copy( it->getName() ) || match_copy_file_in_dir( setup.current_package(), 
+									   it->getFullPathName() ) ) &&  
+		    CppDir::EFILE::REGULAR )
 	      {
 		  if( !setup.unstow() )
 		  {
 		      // copy instead of linking
 		      Link link( it->getFullPathName(), CppDir::concat_dir( target->getFullPathName(), it->getName() ) );		  
-		      link.action = Link::COPY;
+		      link.action = Link::ACTION::COPY;
 		      links->push_back( link );
 		  }		  
 
 		  // a copied file won't be unstowed
 		  continue;
-	      }
+	      } 
 
-	      if( it->getType() == Node::FILE || it->getType() == Node::UNKNOWN )
+	      if( it->getType() == Node::TYPE::FILE || it->getType() == Node::TYPE::UNKNOWN )
 		{
-		  ERROR( "file %s already exist\n", it->getFullPathName() );
-
-		  if( setup.force() )
+		    if( match_ignore( it->getName() ) || match_ignore_file_in_dir( base->getFullPathName(), it->getFullPathName() ) )
 		    {
-		      OUT(0) << "ignoring cause XStow is in force mode\n"
-			"The file will be skipped";
-		      continue;
+			DEBUG( OUT(2)( "file %s ignored\n", it->getFullPathName() ) );
+			continue;
+		    } else {
+			
+			ERROR( "file %s already exist\n", it->getFullPathName() );		  
+			
+			if( setup.force() )
+			{
+			    OUT(0) << "ignoring cause XStow is in force mode\n"
+				"The file will be skipped";
+			    continue;
+			} else {			
+			    return false;
+			}
 		    }
-		  else
-		    return false;
-		}
+		  }
 
-	      if( it->getType() == Node::DIR )
-		{
-		    if( match_protect_dirs( ::getName( *dit ) ) )
+	      if( it->getType() == Node::TYPE::DIR )
+		{		 
+		  if( match_protect_dirs( ::getName( *dit ) ) )
 		    {
 			ERROR( "directory %s is protected\n", ::getName( *dit ) );
 
@@ -428,6 +493,8 @@ bool Tree::check( Ref<Node> target, Ref<Node> base, Ref<vec_link> links )
 			  
 			  CppDir::File f( path );
 			  
+			  DEBUG( OUT(3) << "file: " << path << " is valid? " << (!f ? "NO\n" : "YES\n") );
+
 /*
 			  if( ( !f || path.find( working->getFullPathName() ) != 0 ) &&
 			      ( path.find( target->getRoot()->getFullPathName() ) == 0 ) )
@@ -440,7 +507,7 @@ bool Tree::check( Ref<Node> target, Ref<Node> base, Ref<vec_link> links )
 			      // let's see if it is part of an other stow directory
 			      bool other_stow = false;		  
 
-			      OUT(1) << "cecking if path is a stow dir\n";
+			      OUT(1) << "checking if path is a stow dir\n";
 
 			      if( match_stow_dirs( path ) )
 				{
@@ -496,14 +563,14 @@ bool Tree::check( Ref<Node> target, Ref<Node> base, Ref<vec_link> links )
 			      
 			      Link link;
 			      
-			      link.action = Link::REMOVE;
+			      link.action = Link::ACTION::REMOVE;
 			      link.origin = wpath;
 			      
 			      links->push_back( link );
 			      
 			      if( !checkDirs( *it.it, new Node( f ), wpath, links ) )
 				{			  
-				  OUT(0) << "resolving dendencies failed" << std::endl;
+				  OUT(0) << "resolving dependencies failed" << std::endl;
 				  
 				  if( setup.force() )
 				    {
@@ -520,7 +587,7 @@ bool Tree::check( Ref<Node> target, Ref<Node> base, Ref<vec_link> links )
 
 		    }
 
-		  if( dit->get_type() == CppDir::File::DIR )
+		  if( dit->get_type() == CppDir::EFILE::DIR )
 		    {
 			if( match_copy( dit->get_name() ) )
 			{
@@ -528,7 +595,7 @@ bool Tree::check( Ref<Node> target, Ref<Node> base, Ref<vec_link> links )
 			    {
 				// copy instead of linking
 				Link link( it->getFullPathName(), CppDir::concat_dir( target->getFullPathName(), it->getName() ) );		  
-				link.action = Link::COPY_DIR;
+				link.action = Link::ACTION::COPY_DIR;
 				links->push_back( link );				
 			    }
 
@@ -568,6 +635,14 @@ bool Tree::check( Ref<Node> target, Ref<Node> base, Ref<vec_link> links )
 	{
 	  if( !setup.unstow() )
 	    {
+	      if( match_ignore( it->getName() ) )
+		{
+		  OUT(1)( "%s %s ignored\n", 
+			  it->getType() == Node::TYPE::FILE ? "File" : "Directory" , 
+			  it->getFullPathName() );
+		  continue;
+		} 
+
 		if( !match_protect_targets( CppDir::concat_dir( target->getFullPathName(), it->getName() )  ) )
 		{
 		    ERROR( "not allowed installing file %s there\n", CppDir::concat_dir( target->getFullPathName(), it->getName() ) );
@@ -582,17 +657,19 @@ bool Tree::check( Ref<Node> target, Ref<Node> base, Ref<vec_link> links )
 		    }
 		}
 
-		if( it->getType() == Node::FILE && match_copy( it->getName() ) )
+		if( it->getType() == Node::TYPE::FILE && 
+		    ( match_copy( it->getName() ) || match_copy_file_in_dir( setup.current_package(), 
+									     it->getFullPathName() ) ) )
 		{
 		    // copy instead of linking
 		    Link link( it->getFullPathName(), CppDir::concat_dir( target->getFullPathName(), it->getName() ) );
-		    link.action = Link::COPY;
+		    link.action = Link::ACTION::COPY;
 		    links->push_back( link );
 
-		} else if( it->getType() == Node::DIR && match_copy( it->getName() ) ) {
+		} else if( it->getType() == Node::TYPE::DIR && match_copy( it->getName() ) ) {
 
 		    Link link( it->getFullPathName(), CppDir::concat_dir( target->getFullPathName(), it->getName() ) );
-		    link.action = Link::COPY_DIR;
+		    link.action = Link::ACTION::COPY_DIR;
 		    links->push_back( link );
 
 		} else {
@@ -614,27 +691,41 @@ bool Tree::check( Ref<Node> target, Ref<Node> base, Ref<vec_link> links )
 
 std::ostream& operator<<( std::ostream& out, const Tree::Link &link )
 {
+    BT;
+
   switch( link.action )
     {
-    case Tree::Link::SET_LINK:
+    case Tree::Link::ACTION::SET_LINK:
       out << "SET_LINK: " << link.origin << " => " << link.target;
       break;
       
-    case Tree::Link::REMOVE:
+    case Tree::Link::ACTION::REMOVE:
       out << "REMOVE: " << link.origin;
       break;
       
-    case Tree::Link::MKDIR:
+    case Tree::Link::ACTION::MKDIR:
       out << "MKDIR: " << link.target;
       break;
 
-    case Tree::Link::COPY:
+    case Tree::Link::ACTION::COPY:
       out << "COPY: " << link.origin;
       break;
 
-    case Tree::Link::COPY_DIR:
+    case Tree::Link::ACTION::COPY_DIR:
       out << "COPY: " << link.origin;
       break;
+
+#ifdef CAN_USE_INI
+    case Tree::Link::ACTION::EXEC:
+      out << "EXEC: " << command( link.exec, link.origin, link.target );
+      // out << "EXEC: " << link.origin << " => " << link.target << " => " << link.exec;
+      break;
+#endif
+
+	case Tree::Link::ACTION::FIRST__:
+	case Tree::Link::ACTION::LAST__:
+		break;
+
     }
   return out;
 }
@@ -658,6 +749,7 @@ std::ostream& operator<<( std::ostream& out, Ref<Tree::vec_link> links )
 
 bool Tree::checkDirs( Ref<Node> a, Ref<Node> b, std::string wpath, Ref<vec_link> links)
 {
+    BT;
 
   DEBUG( OUT(3)( "wpath: %s\n", wpath ) );
   DEBUG( OUT(3)( "a->getFullPathName(): %s\n", a->getFullPathName() ) );
@@ -665,7 +757,7 @@ bool Tree::checkDirs( Ref<Node> a, Ref<Node> b, std::string wpath, Ref<vec_link>
 
   Link link;
 
-  link.action = Link::MKDIR;
+  link.action = Link::ACTION::MKDIR;
   link.target = wpath;
   
   links->push_back( link );
@@ -682,9 +774,12 @@ bool Tree::checkDirs( Ref<Node> a, Ref<Node> b, std::string wpath, Ref<vec_link>
 	{
 	  DEBUG( OUT(3)( "ait->getName() == bit->getName() => %s == %s\n",  ait->getName(), bit->getName() ) );
 
+	  bool failed = true;
+
 	  if( ait->getName() == bit->getName() )
 	    {
-	      if( ait->getType() == Node::DIR && bit->getType() == Node::DIR )
+
+	      if( ait->getType() == Node::TYPE::DIR && bit->getType() == Node::TYPE::DIR )
 	      {
 		if( !checkDirs( *ait.it, *bit.it, 
 				CppDir::concat_dir( wpath, ait->getName() ), links ) )
@@ -692,14 +787,33 @@ bool Tree::checkDirs( Ref<Node> a, Ref<Node> b, std::string wpath, Ref<vec_link>
 		
 		Link link;
 		
-		link.action = Link::MKDIR;
+		link.action = Link::ACTION::MKDIR;
 		link.target = CppDir::concat_dir( wpath, ait->getName() );
 
 		links->push_back( link );
 
-	      } else {
+		failed = false;
 
-		OUT(0)( "File %s and file %s are not of the same type", 
+	      } else if( ait->getType() == Node::TYPE::FILE && bit->getType() == Node::TYPE::FILE ) { 
+			 		  
+		  if( match_copy( ait->getName() ) || match_copy_file_in_dir( setup.current_package(), 
+									      ait->getFullPathName() ) ) 
+		  {
+		      DEBUG( OUT(2)( "file %s will be copied\n", ait->getFullPathName() ) );
+		      failed = false;
+
+		  } else if( match_ignore( ait->getName() ) || 
+			     match_ignore_file_in_dir( base->getFullPathName(), ait->getFullPathName() ) ) {
+
+		      DEBUG( OUT(2)( "file %s will be ignored\n", ait->getFullPathName() ) );
+		      failed = false;
+		  }
+
+	      } 
+
+	      if( failed ) {
+
+		OUT(0)( "File %s and file %s are not of the same type\n", 
 			ait->getFullPathName(), bit->getFullPathName() );
 
 		return false;
@@ -725,6 +839,7 @@ bool Tree::checkDirs( Ref<Node> a, Ref<Node> b, std::string wpath, Ref<vec_link>
 
       if( !found )
 	{
+          DEBUG( OUT(3) << "add link: " << bit->getFullPathName() << " => " << CppDir::concat_dir( wpath, bit->getName() )  << '\n' );
 	  links->push_back( Link( bit->getFullPathName(), CppDir::concat_dir( wpath, bit->getName() ) ) );
 	} 	    
     }
@@ -734,6 +849,8 @@ bool Tree::checkDirs( Ref<Node> a, Ref<Node> b, std::string wpath, Ref<vec_link>
 
 void Tree::unify_links( Ref<vec_link> &l )
 {
+    BT;
+
     Ref<vec_link> ll = new vec_link;
 
     for( vec_link::iterator it = l->begin(); it != l->end(); ++it )
@@ -788,6 +905,8 @@ bool operator<( const Tree::Link &a, const Tree::Link &b )
 
 bool Tree::exec_list()
 {
+    BT;
+
   unify_links( links );
 
   DEBUG( OUT(3) << '\n' << links << '\n' );
@@ -815,7 +934,7 @@ bool Tree::exec_list()
 
       switch( it->action )
 	{
-	case Link::REMOVE:
+	case Link::ACTION::REMOVE:
 	  {
 	    CppDir::File file( it->origin );
 	    
@@ -825,7 +944,7 @@ bool Tree::exec_list()
 		return false;
 	      }
 
-	    if( !file.is_link() && file.get_type() == CppDir::File::DIR )
+	    if( !file.is_link() && file.get_type() == CppDir::EFILE::DIR )
 	      {
 		// remove directories
 		
@@ -847,7 +966,7 @@ bool Tree::exec_list()
 		    if( rmdir( it->origin.c_str() ) !=  0 )
 		      {
 			ERROR( "unlink: errno: %d\n", errno );
-			eout << strerror( errno ) << '\n';
+			eout << std::strerror( errno ) << '\n';
 			eout( "file: %s\n", it->origin );
 			return false;
 		      }
@@ -862,7 +981,7 @@ bool Tree::exec_list()
 		if( unlink( it->origin.c_str() ) != 0 )
 		  {
 		    ERROR( "unlink: errno: %d\n", errno );
-		    eout << strerror( errno ) << '\n';
+		    eout << std::strerror( errno ) << '\n';
 		    return false;
 		  }
 	      }
@@ -895,7 +1014,7 @@ bool Tree::exec_list()
 			  {
 			    Link link;
 			    
-			    link.action = Link::REMOVE;
+			    link.action = Link::ACTION::REMOVE;
 			    link.origin = file.get_path();
 			    
 			    l.push_back( link );
@@ -907,7 +1026,7 @@ bool Tree::exec_list()
 	  } // Link::Remove
 	  break;
 
-	case Link::SET_LINK:
+	case Link::ACTION::SET_LINK:
 	  {
 	    errno = 0;
 
@@ -916,52 +1035,71 @@ bool Tree::exec_list()
 		ERROR( "symlink: errno: %d\n", errno );
 		OUT(1) << "origin: " << it->origin << '\n';
 		OUT(1) << "target: " << it->target << '\n';
-		eout << strerror( errno ) << '\n';
+		eout << std::strerror( errno ) << '\n';
 		
 		return false;
 	      }
 	  }
 	break;
 
-	case Link::MKDIR:
+	case Link::ACTION::MKDIR:
 	  {
 	    errno = 0;
 
 	    if( mkdir( it->target.c_str(), 0755 ) != 0 )
 	      {
 		ERROR( "mkdir: %s\nerrno: %d\n", it->target.c_str(), errno );
-		eout << strerror( errno ) << '\n';
+		eout << std::strerror( errno ) << '\n';
 
 		return false;
 	      }
 	  }
 	break;
 
-	case Link::COPY:
+	case Link::ACTION::COPY:
 	  {
 	      errno = 0;
 
 	      if( !file_copy( it->origin, it->target ) )
 	      {
 		  ERROR( "file_copy: %s\nerrno: %d\n", it->target.c_str(), errno );
-		  eout << strerror( errno ) << '\n';
+		  eout << std::strerror( errno ) << '\n';
 	      }
 	  }
 	break;
 
-	case Link::COPY_DIR:
+	case Link::ACTION::COPY_DIR:
 	  {
 	      errno = 0;
 
 	      if( !dir_copy( it->origin, it->target ) )
 	      {
 		  ERROR( "dir_copy: %s\nerrno: %d\n", it->target.c_str(), errno );
-		  eout << strerror( errno ) << '\n';
+		  eout << std::strerror( errno ) << '\n';
 	      }
 
 	  }
 	break;
 
+#ifdef CAN_USE_INI
+       case Link::ACTION::EXEC:
+         {
+	     std::string com = command( it->exec, it->origin, it->target );
+
+	     int ret = system( com.c_str() );
+
+	     if( ret == -1 )
+		 ERROR( "system() failed\n" );		 
+	     if( ret != 0 )
+		 ERROR( "exit status of command %s was %d\n", com, ret );
+         }
+	 break;
+#endif
+
+	case Link::ACTION::FIRST__:
+	case Link::ACTION::LAST__:
+		break;
+
 	} // switch
     } // for
 
@@ -970,9 +1108,11 @@ bool Tree::exec_list()
 
 void make_relative_links( Ref<Tree::vec_link> links )
 {
+    BT;
+
   for( Tree::vec_link::iterator it = links->begin(); it != links->end(); ++it )
     {
-      if( it->action == Tree::Link::SET_LINK )
+      if( it->action == Tree::Link::ACTION::SET_LINK )
 	{
 	  std::string path, name;
 
@@ -987,6 +1127,7 @@ void make_relative_links( Ref<Tree::vec_link> links )
 
 bool match_ignore( const std::string &file )
 {
+    BT;
 
 #ifdef HAVE_REGEX_H
   for( unsigned int i = 0; i < setup.matches.ignore_regex.size(); ++i )
@@ -1001,7 +1142,13 @@ bool match_ignore( const std::string &file )
   for( unsigned int i = 0; i < setup.matches.ignore.size(); ++i )
     {
       if( fnmatch( setup.matches.ignore[i].c_str(), file.c_str(), FNM_PATHNAME ) == 0 )
-	return true;
+	{
+	  return true;
+	}
+      else
+	{
+	  DEBUG( OUT(3)( "fnmatch failed: '%s' != '%s'",  setup.matches.ignore[i], file ) );
+	}
     }
 #endif
 
@@ -1010,6 +1157,7 @@ bool match_ignore( const std::string &file )
 
 bool match_copy( const std::string &file )
 {
+    BT;
 
 #ifdef HAVE_REGEX_H
   for( unsigned int i = 0; i < setup.matches.copy_regex.size(); ++i )
@@ -1033,6 +1181,7 @@ bool match_copy( const std::string &file )
 
 bool match_auto_traverse( const std::string &file )
 {
+    BT;
 
 #ifdef HAVE_REGEX_H
   for( unsigned int i = 0; i < setup.traversable_links.add_if_targets_regex.size(); ++i )
@@ -1057,6 +1206,8 @@ bool match_auto_traverse( const std::string &file )
 
 bool match_stow_dirs( const std::string &file )
 {
+    BT;
+
   for( unsigned int aa=0; aa < setup.stow_dirs.dirs.size(); ++aa )
     {
       if( CppDir::is_in_dir( file, setup.stow_dirs.dirs[aa] ) )
@@ -1068,6 +1219,9 @@ bool match_stow_dirs( const std::string &file )
 
 bool match_auto_stow_dirs( const std::string &file )
 {
+    BT;
+
+    DEBUG( OUT(3) << "file: " << file << '\n' );
 
 #ifdef HAVE_REGEX_H
   for( unsigned int i = 0; i < setup.stow_dirs.auto_add_dirs_regex.size(); ++i )
@@ -1092,6 +1246,7 @@ bool match_auto_stow_dirs( const std::string &file )
 
 bool match_protect_dirs( const std::string &file )
 {
+    BT;
 
     for( unsigned int aa=0; aa < setup.protect_dirs.dirs.size(); ++aa )
     {
@@ -1104,9 +1259,12 @@ bool match_protect_dirs( const std::string &file )
 
 bool match_auto_protect_dirs( const std::string &file )
 {
+    BT;
+
+    DEBUG( OUT(3) << "file: " << file << '\n' );
 
 #ifdef HAVE_REGEX_H
-  for( unsigned int i = 0; i < setup.stow_dirs.auto_add_dirs_regex.size(); ++i )
+  for( unsigned int i = 0; i < setup.protect_dirs.auto_add_dirs_regex.size(); ++i )
     {
       if( regexec( &setup.protect_dirs.auto_add_dirs_regex.regex(i),
 		   file.c_str(), 0, 0, 0 ) == 0 )
@@ -1128,6 +1286,8 @@ bool match_auto_protect_dirs( const std::string &file )
 
 bool match_protect_targets( const std::string &file )
 {
+    BT;
+
     if( setup.protect_dirs.targets.size() == 0 )
 	return true; // value not set
 
@@ -1140,96 +1300,76 @@ bool match_protect_targets( const std::string &file )
   return false;
 }
 
-#include <cstdio>
-
-bool file_copy( std::string origin, std::string target )
+#ifdef CAN_USE_INI
+bool match_exec( const std::string &file, std::string &exec )
 {
-    FILE *o,*t;       
+    BT;
 
-    DEBUG( OUT( 3 )( "file_copy: %s => %s\n", origin, target ) );
-
-    if( ( o = std::fopen( origin.c_str(), "rb" ) ) == NULL )
-	return false;
-
-    DEBUG( OUT( 3 ) << "opening origin succeeded\n"  );
-
-    if( ( t = std::fopen( target.c_str(), "w+b" ) ) == NULL )
-	return false;
-
-    DEBUG( OUT( 3 ) << "opening target succeeded\n"  );
-
-    const int SIZE = 1024*1024;
-    char buffer[SIZE];
-
-#ifdef NO_STD_FEOF
-    if( ferror( o ) || ferror( t ) )
-#else
-    if( std::ferror( o ) || std::ferror( t ) )
-#endif
-	return false;
-
-    DEBUG( OUT( 3 ) << "no ferror\n"  );
-
-    do
+    for( unsigned i = 0; i < setup.exec.size(); ++i )
     {
-	size_t rsize = std::fread( buffer, 1, SIZE, o );
-	std::fwrite( buffer, 1, rsize, t );
-
-	DEBUG( OUT(3)( "wrote %d: bytes error: %d\n", rsize, std::ferror( o ) ) );
-
-#ifdef NO_STD_FEOF
-    } while( !feof( o ) && !ferror( o ) && !ferror( t ) );
-#else
-    } while( !std::feof( o ) && !std::ferror( o ) && !std::ferror( t ) );
-#endif
-
-    std::fclose( o );
-    std::fclose( t );
-
-    return true;
-}
-
-bool dir_copy( std::string origin, std::string target )
-{
-    CppDir::Directory dir( target );
-    
-    if( !dir )
-    {
-	// directory does not exist, create it
-	if( mkdir( target.c_str(), 0777 ) != 0 )
-	    return false;
-
-	dir = CppDir::Directory( target );
-    }
-
-    CppDir::Directory odir( origin );    
-
-    CppDir::Directory::file_list fl = odir.get_files();
-
-    for( CppDir::Directory::file_list_it it = fl.begin(); it != fl.end(); ++it )
-    {
-	if( it->get_name() == "." ||
-	    it->get_name() == ".." )
-	    continue;
-
-	std::string no = CppDir::concat_dir( origin, it->get_name() );
-	std::string nt = CppDir::concat_dir( target, it->get_name() );
-
-	if( it->get_type() == CppDir::File::DIR )
+	if( setup.exec[i].match() == file )
 	{
-	    if( !dir_copy( CppDir::concat_dir( origin, it->get_name() ),
-			   CppDir::concat_dir( target, it->get_name() ) ) )
-		return false;
-
-	} else {
-
-	    OUT(1)( "COPY: %s => %s\n", no, nt );
-	
-	    if( !file_copy(  no, nt  ) )
-		return false;
+	    if( setup.unstow() && !setup.exec[i].exec_unstow().empty() )
+		exec = setup.exec[i].exec_unstow();
+	    else
+		exec = setup.exec[i].exec();
 
+	    return true;
 	}
+
+#ifdef HAVE_FNMATCH_H
+	if( fnmatch(  setup.exec[i].match().c_str(), file.c_str(), 
+		      FNM_PATHNAME | FNM_PERIOD ) == 0 )
+	{
+	    exec = setup.exec[i].exec();
+	    return true;
+	}
+#endif
     }
-   
-    return true;
+
+    return false;
 }
+
+std::string command( std::string exec, const std::string &source, const std::string &target )
+{
+    exec = substitude( exec, "%s", '\"' + source + '\"' );
+    exec = substitude( exec, "%t", '\"' + target + '\"' );
+
+    return exec;
+}
+#endif
+
+bool match_ignore_file_in_dir( const std::string working_dir, const std::string &file )
+{
+    BT;
+
+    DEBUG( OUT(4) << "working dir: " << working_dir << " file: " << file << '\n' );
+
+    for( unsigned i = 0; i < setup.matches.ignore_file_in_dir.size(); i++ )
+    {
+	std::string f = CppDir::concat_dir( working_dir, setup.matches.ignore_file_in_dir[i] );
+
+	if( f == file )
+	    return true;
+    }
+
+    return false;
+}
+
+bool match_copy_file_in_dir( const std::string working_dir, const std::string &file )
+{
+    BT;
+
+    DEBUG( OUT(4) << "working dir: " << working_dir << " file: " << file << '\n' );
+
+    for( unsigned i = 0; i < setup.matches.copy_file_in_dir.size(); i++ )
+    {
+	std::string f = CppDir::concat_dir( working_dir, setup.matches.copy_file_in_dir[i] );
+
+	if( f == file )
+	    return true;
+    }
+
+    return false;
+}
+
